#!/usr/bin/env python3
#
# Copyright 2020 David A. Greene
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <https://www.gnu.org/licenses/>.
#

"""A plugin to add an 'init' command to git-project.  The init command does
initialization of a workarea,

Summary:

git-project init

"""
from git_project import Git, RunnableConfigObject, Plugin
from git_project import add_top_level_command, Project, GitProjectException

from git_project_core_plugins.common import add_plugin_version_argument

import getpass

def command_init(git, gitproject, project, clargs):
    """Implement git-project init."""
    pass

class InitPlugin(Plugin):
    """
    The init command initializes project state.

    Summary:

      git <project> init

    Basic config entries are added to name the project and default branches.
    Plugins may add options to enhance functionality.  For example the worktree
    command adds a --worktree option to convert an existing local clone to a
    ``worktree layout.''

    See also:

      worktree

    """
    def __init__(self):
        super().__init__('init')

    def add_arguments(self,
                      git,
                      gitproject,
                      project,
                      parser_manager,
                      plugin_manager):
        """Add arguments for 'git project init.'"""
        # init
        init_parser = add_top_level_command(parser_manager,
                                            'init',
                                            'init',
                                            help='Initialize project')

        add_plugin_version_argument(init_parser)

        init_parser.set_defaults(func=command_init)



#!/usr/bin/env python3
#
# Copyright 2020 David A. Greene
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <https://www.gnu.org/licenses/>.
#

from .artifact import Artifact, ArtifactPlugin
from .branch import BranchPlugin
from .run import RunPlugin
from .clone import ClonePlugin
from .common import add_plugin_version_argument
from .config import ConfigPlugin
from .help import Help, HelpPlugin
from .init import InitPlugin
from .worktree import Worktree, WorktreePlugin



#!/usr/bin/env python3
#
# Copyright 2020 David A. Greene
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <https://www.gnu.org/licenses/>.
#

"""A plugin to add a 'config' command to git-project.  The config command sets
project-wide git configuration values and prints their values to stdout.

Summary:

git-project config <key> [--unset] [<value>]

"""

from git_project import ConfigObject, SubstitutableConfigObject, Plugin
from git_project import run_command_with_shell, add_top_level_command

from git_project_core_plugins.common import add_plugin_version_argument

import argparse
import re

class Artifact(SubstitutableConfigObject):
    @classmethod
    def _split_ident(cls, ident):
        parts = ident.rsplit('.', 1)

        ident = parts[-1]

        subsection = cls.subsection()
        if len(parts) > 1:
            subsection += '.' + '.'.join(parts[:-1])

        return subsection, ident

    @classmethod
    def get(cls, git, project_section, ident, **kwargs):
        """Factory to construct an Artifact object.

        git: An object to query the repository and make config changes.

        project_section: the active project section.

        ident: The subsection under <project>.artifact where this Artifact will
               live.

        **kwargs: Keyword arguments of property values to set upon
                  construction.

        """
        subsection, ident = cls._split_ident(ident)

        return super().get(git,
                           project_section,
                           subsection,
                           ident,
                           **kwargs)

    @classmethod
    def exists(cls,
               git,
               project_section,
               ident):
        """Return whether an existing git config exists for the Artifact.

        cls: The derived class being checked.

        git: An object to query the repository and make config changes.

        project_section: git config section of the active project.

        ident: The subsection under <project>.artifact where this Artifact lives.

        """
        subsection, ident = cls._split_ident(ident)

        return super().exists(git, project_section, subsection, ident)

    def __init__(self,
                 git,
                 project_section,
                 subsection,
                 ident,
                 **kwargs):
        """Artifact construction.

        cls: The derived class being constructed.

        git: An object to query the repository and make config changes.

        project_section: git config section of the active project.

        subsection: An arbitrarily-long subsection appended to project_section

        ident: The name of this specific Artifact.

        **kwargs: Keyword arguments of property values to set upon construction.

        """
        super().__init__(git,
                         project_section,
                         subsection,
                         ident,
                         **kwargs)

    @staticmethod
    def subsection():
        """ConfigObject protocol subsection."""
        return 'artifact'

    @classmethod
    def get_managing_command(cls):
        return 'artifact'

    def rm(self, git, project, clargs, string):
        """Remove the paths associated with this Artifact.

        git: An object to query the repository and make config changes.

        project: The currently active Project.

        clargs: Command-line arguments

        string: The string on which to perform substitution.

        """
        for path in self.iter_multival('path'):
            path = self.substitute_value(git, poject, clargs, path)

def command_artifact_add(git, gitproject, project, clargs):
    """Implement git-project artifact add."""
    ident = clargs.subsection
    path = clargs.path

    artifact = Artifact.get(git, project.get_section(), ident)
    artifact.add_item('path', path)

def command_artifact_rm(git, gitproject, project, clargs):
    """Implement git-project artifact rm."""
    ident = clargs.subsection

    artifact = Artifact.get(git, project.get_section(), ident)

    if hasattr(clargs, 'path') and clargs.path:
        path = clargs.path
        artifact.rm_item('path', path)
    else:
        artifact.rm_items('path')

class ArtifactPlugin(Plugin):
    """
    The artifact command adds or removes associations between git config objects
    and file-system objects.

    Summary:

      git <project> artifact add <subsection> <path>
      git <project> artifact rm <subsection> [<path>]

    <subsection> is a git config section which will appear under the
    <project>.artifact section.  Artifacts look up objects associated with
    <subsection> and perform substitutions on paths to yield the final
    associated file-system object.  The ``artifact rm'' command simply removes
    an artifact association, it does not remove the artifact itself.  Multiple
    artifact paths may be associated under a single <subsection> and the option
    <path> argument to ``artifact rm'' allows us to remove a single association
    rather than all of them at once.

    For example:

      git <project> artifact add worktree.myworktree /path/to/artifact

    Presumably, /path/to/artifact is in some way created in association with
    myworktree, for example by the ``run'' command.  When we delete myworktree,
    the artifact association causes /path/to/artifact to also be removed.
    Substitutions can make artifact associations easier to manage:

      git <project> artifact add worktree /path/to/{worktree}/artifact

    Notice that we've added the artifact under the more general ``worktree''
    subsection instead of naming a worktree explicitly as before.  Because the
    {worktree} substitution appears in the artifact path, deleting any worktree
    will cause the worktree's name to be substituted into the artifact path,
    forming a unique artifact path to remove.

    We may make this even more general:

      git <project> config srcdir "{path}"
      git <project> config builddir "{srcdir}/build/{worktree}"
      git <project> config make "make -C {srcdir} BUILDDIR={builddir} {build}"
      git <project> run --make-alias build
      git <project> add build debug "{make}"
      git <project> add build release "{make}"
      git <project> add build check "{make}"
      git <project> artifact add worktree "{builddir}"

    We've added a single artifact association that will handle any worktree and
    all of our different build types.  When we delete the worktree, all
    artifacts related to debug, release and check builds will also be removed.

    The worktree plugin also modifies the clone and init commands, adding a
    --worktree option to both.  With --worktree, clone will create a ``worktree
    layout'' as so:

      clonedir
        .git
         master

    Here, ``master'' is a worktree created from the master branch.  ``clonedir''
    becames a bare repository, though with refspecs that make it operate like a
    regular clone for fetch and push operations.  That is, the cloned repository
    will still have refs/heads and refs/remotes namespaces.

    With --worktree, init will take an existing local clone and convert it to a
    bare repository, removing all checked out files and creating a master
    worktree:

      clonedir
        .git
         master

    Conversion will abort if the workarea is not in a clean state.  Note that
    all files in clonedir will be deleted so if there are important files not
    part of the underlying repository, the user must take care to preserve them.
    If the workarea hd a branch other than master checked out, no worktree for
    it will be created automatically, though the user may easily create one
    after conversion.

    See also:

      clone
      config
      init
      run
      worktree

    """

    def __init__(self): super().__init__('artifact')

    def add_arguments(self,
                      git,
                      gitproject,
                      project,
                      parser_manager,
                      plugin_manager):
        """Add arguments for 'git-project artifact.'"""

        artifact_parser = add_top_level_command(parser_manager,
                                                'artifact',
                                                'artifact',
                                                help='Manipulate artifacts')

        add_plugin_version_argument(artifact_parser)

        artifact_subparser = parser_manager.add_subparser(artifact_parser,
                                                          'artifact-command',
                                                          help='artifact commands')

        # add
        add_parser = parser_manager.add_parser(artifact_subparser,
                                               'add',
                                               'artifact-add',
                                               help='Add an artifact',
                                               epilog = """
The <subsection> argument is appended to the <project>.artifact section to form
the final git config section that will hold the artifact path.
""",
                                               formatter_class = argparse.RawDescriptionHelpFormatter)

        add_parser.add_argument('subsection',
                                help='Subsection under which to add the artifact')
        add_parser.add_argument('path',
                                help='Artifact path, may use substitutions')

        add_parser.set_defaults(func=command_artifact_add)

        # rm
        rm_parser = parser_manager.add_parser(artifact_subparser,
                                              'rm',
                                              'artifact-rm',
                                              help='Remove an artifact path',
                                              epilog = """
The <subsection> argument is appended to the <project>.artifact section to form
the final git config section that will hold the artifact path.
""",
                                              formatter_class =
                                              argparse.RawDescriptionHelpFormatter)

        rm_parser.add_argument('subsection',
                               help='Subsection under which to add the artifact')
        rm_parser.add_argument('path', nargs='?',
                               help='Artifact path, may use substitutions')

        rm_parser.set_defaults(func=command_artifact_rm)

    def add_class_hooks(self, git, project, plugin_manager):
        # Enhance ConfigObject rm to also remove artifacts.
        config_object_rm = ConfigObject.rm

        def artifact_rm(self):
            # See if there is any artifact associated with this ConfigObject.
            artifact = None
            if Artifact.exists(self._git,
                               self._project_section,
                               self._subsection + '.' + self._ident):
                artifact = Artifact.get(self._git,
                                        self._project_section,
                                        self._subsection + '.' + self._ident)
            elif Artifact.exists(self._git,
                                 self._project_section,
                                 self._subsection):
                artifact = Artifact.get(self._git,
                                        self._project_section,
                                        self._subsection)

            if artifact:
                for path in artifact.iter_multival('path'):
                    fullpath = artifact.substitute_value(self._git, project, path)
                    run_command_with_shell(f'rm -rf {fullpath}',
                                           show_command=True)

            config_object_rm(self)

        ConfigObject.rm = artifact_rm



#!/usr/bin/env python3
#
# Copyright 2020 David A. Greene
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <https://www.gnu.org/licenses/>.
#

from git_project import ConfigObject, Git, Plugin, Project
from git_project import ScopedConfigObject
from git_project import add_top_level_command, GitProjectException

from git_project_core_plugins.common import add_plugin_version_argument

import argparse
import os
from pathlib import Path
import shutil
import urllib

# Take a path and normalize it to the current working directory.  If the current
# directory is a bare repository, place this path inside it.  If not, place this
# path outside it.
def normalize_path(git, path):
    """Find an appropriate repository-relative path.  Given a path, if it is not an
    absolute path, place it in the current directory if the repository is a bare
    repository or place it one level above the repository root directory if it
    is not a bare repository.

    path: A string path

    """
    path = Path(path).expanduser()

    if not path.is_absolute() and path.parts[0] != '..':
        # If this is a bare repository, put this in a directory below it.
        # Otherwise put it in the directory above it.
        if git.is_bare_repository():
            path = Path.cwd() / path
        else:
            path = Path(git.get_working_copy_root()) / path

    path = path.resolve()

    return str(path)

# Determine a path and committish from args.
def get_name_path_and_refname(git, gp, clargs):
    """Given a Project and worktree command-line arguments <name-or-path> and
    <committish>, determine an appropriate worktree name, a path based on the
    name and refname based on the name.  If one or the other of <name-or-path>
    or <committish> is not provided, figure it out from the provided value.  If
    neither <name-or-path> nor <committish> is provided, raise an exception.

    """
    name = str(Path(clargs.path).name)
    path = normalize_path(git, clargs.path)
    refname = git.committish_to_refname('HEAD')
    if hasattr(clargs, 'committish') and clargs.committish:
        refname = git.committish_to_refname(clargs.committish)

    return name, path, refname

# worktree add
def command_worktree_add(git, gitproject, project, clargs):
    """Implement git-project worktree add."""
    name, path, refname = get_name_path_and_refname(git, gitproject, clargs)

    branch = git.refname_to_branch_name(refname)
    branch_point = refname

    if not git.committish_exists(branch_point):
        raise GitProjectException(f'Branch point {branch_point} does not exist for worktree add')

    # Either use the branch the user gave us or create a branch (if needed)
    # named after the given name.
    if hasattr(clargs, 'branch') and clargs.branch:
        branch = clargs.branch
        git.create_branch(branch, branch_point)
    elif name != branch:
        branch = name
        if not git.committish_exists(name):
            git.create_branch(branch, branch_point)

    worktree = Worktree.get(git,
                            project,
                            name,
                            path=path,
                            committish=branch)
    worktree.add()

    return worktree

def command_worktree_rm(git, gitproject, project, clargs):
    """Implement git-project worktree rm."""
    name = clargs.name
    worktree = Worktree.get(git, project, name)

    if not project.branch_is_merged(worktree.committish) and not clargs.force:
        raise GitProjectException(f'Worktree branch {worktree.committish} is not merged, use -f to force')

    worktree.rm()

class Worktree(ScopedConfigObject):
    """A ScopedConfigObject to manage worktree git configs."""
    class Path(ConfigObject):
        """A ConfigObject to manage worktree paths.  Each worktree config section has an
        associated worktreepath config section to allow fast mapping from a
        worktree path to its Worktree ConfigObject.

        """

        def __init__(self,
                     git,
                     project_section,
                     subsection,
                     ident,
                     **kwargs):
            """Path construction.

            cls: The derived class being constructed.

            git: An object to query the repository and make config changes.

            project_section: git config section of the active project.

            subsection: An arbitrarily-long subsection appended to project_section

            ident: The name of this specific Build.

            **kwargs: Keyword arguments of property values to set upon construction.

            """
            super().__init__(git,
                             project_section,
                             subsection,
                             ident,
                             **kwargs)

        @classmethod
        def subsection(cls):
            """ConfigObject protocol subsection."""
            return 'worktreepath'

        @classmethod
        def get(cls, git, project_section, path, **kwargs):
            """Factory to construct a worktree Path object.

            git: An object to query the repository and make config changes.

            project_section: git config section of the active project.

            path: The path to reference this Path..

            **kwargs: Keyword arguments of property values to set upon
                      construction.

            """
            return super().get(git,
                               project_section,
                               cls.subsection(),
                               path,
                               **kwargs)

    def __init__(self,
                 git,
                 project_section,
                 subsection,
                 ident,
                 **kwargs):
        """Worktree construction.

        cls: The derived class being constructed.

        git: An object to query the repository and make config changes.

        project_section: git config section of the active project.

        subsection: An arbitrarily-long subsection appended to project_section

        ident: The name of this specific Build.

        **kwargs: Keyword arguments of property values to set upon construction.

        """
        super().__init__(git,
                         project_section,
                         subsection,
                         ident,
                         **kwargs)
        self._pathsection = self.Path.get(git,
                                          project_section,
                                          self.path,
                                          worktree=ident)

    @staticmethod
    def subsection():
        """ConfigObject protocol subsection."""
        return 'worktree'

    @classmethod
    def get(cls, git, project, name, **kwargs):
        """Factory to construct Worktrees.

        cls: The derived class being constructed.

        git: An object to query the repository and make config changes.

        project: The currently active Project.

        name: Name of the worktree to construct.

        """
        worktree = super().get(git,
                               project.get_section(),
                               cls.subsection(),
                               name,
                               **kwargs)
        project.push_scope(worktree)
        return worktree

    @classmethod
    def get_managing_command(cls):
        return 'worktree'

    @classmethod
    def get_by_path(cls, git, project, path):
        """Given a Project section and a path, get the associated Worktree."""
        pathsection = cls.Path.get(git, project.get_section(), path)
        if hasattr(pathsection, 'worktree'):
            assert pathsection.worktree
            return cls.get(git, project, pathsection.worktree, path=path)
        return None

    def add(self):
        """Create a new worktree"""
        self._git.add_worktree(self.get_ident(), self.path, self.committish)

    def rm(self):
        """Remove a worktree, deleting its workarea, builds and installs."""
        # TODO: Use python utils.
        try:
            shutil.rmtree(self.path)
            shutil.rmtree(self.builddir)
            shutil.rmtree(self.prefix)
            shutil.rmtree(self.installdir)
        except:
            pass

        self._git.prune_worktree(self.get_ident())

        project = Project.get(self._git, self._project_section)

        for branch in project.iterbranches():
            branch_name = self._git.committish_to_refname(branch)
            committish_name = self._git.committish_to_refname(self.committish)
            if branch_name == committish_name:
                break
        else:
            project.prune_branch(self.committish)

        self._pathsection.rm()
        super().rm()

class WorktreePlugin(Plugin):
    """
    The worktree command manages worktrees and connects them to projects.

    Summary:

      git <project> worktree add [-b <branch>] <name-or-path> [<committish>]
      git <project> worktree rm <name-or-path>
      git <project> worktree config <key> [<value>]
      git <project> worktree config [--unset] <key> [<value>]

    ``worktree add'' creates a new git worktree named via <name-or-path> with
    <committish> checked out.  If we pass -b <branch> we'll get a new branch at
    HEAD or <committish> if it is given.  The worktree name is either the given
    name or if name-or-path is a path, the worktree name will be the same as the
    last path component.  If <name-or-path> is a simple name with no directory
    separators, the worktree will be created as a sub-directory of the current
    directory.

    To keep things simple, we'll usually always name worktrees similarly (or
    identically) to the branches they reference, though it is not strictly
    necessary to do so.

    The key idea behind project worktrees is that they are connected to various
    ``artifacts.''  Worktrees are managed together with this artifacts to
    provide a project-level view of various tasks.  For example, a ``run''
    command can create artifacts associated with a worktree.  Removing the
    worktree implicitly removes thee artifacts, making build cleanups easy and
    convenient.  Commands may use the {worktree} substitution to create
    worktree-unique artifacts.  Other substitutions may also referece {worktree}
    in a recursive manner.

    Here is a concrete example:

      git <project> config srcdir "{path}"
      git <project> config builddir "{srcdir}/build/{worktree}"
      git <project> config make "make -C {srcdir} BUILDDIR={builddir} {build}"
      git <project> run --make-alias build
      git <project> add build release "{make}"

    Assuming the build system uses BUILDDIR to determine where build artifacts
    go, each worktree will get a unique set of build artifacts, via the
    {builddir} and, recursively., {worktree} substitutions.  When we delete the
    worktree, we'll also delete the associoated build directory.

    We associate artifacts with worktrees via the artifact commands.

    Another important benefit of worktrees and associated builds is that
    switching to work on a new worktree (by simply editing sources in a
    different worktree directory) will not result in build artifacts from thte
    previous worktree being overwritten.  Thus we avoid the ``rebuild the
    world'' problems of switching branches within the same workarea.  Generally,
    each created branch will have its own worktree and we will rarely, if ever,
    switch branches within a worktree.

    A worktree layers a config scope on top of the global project scope, so that
    configuring a key in the worktree with the same name as a key in the project
    will cause the worktree key's value to override the project key's value:

      git <project> config buildwidth 16
      git <project> worktree config buildwidth 32

    If we are in a worktree configured with buildwidth=32, then wherever
    {buildwidth} apeears (say, in a run command), the value 32 will be
    substituted instead of 16.  If we are outside the worktree (for example a
    worktree without a buildwidth configured), then {buildwidth} will be
    substituted with 16.

    See also:

      artifact
      config
      run

    """

    def __init__(self):
        super().__init__('worktree')

    def initialize(self, git, gitproject, project, plugin_manager):
        """Instantiate a Worktree if we are in a worktree path, providing scoping for
        Project config variables.

        git: A Git object to examine the repository.

        gitproject: A GitProject object to explore and manipulate the active
                    project.

        project: The active Project.

        plugin_manager: The active  PluginManager.

        """
        path = Path.cwd()
        path.resolve()
        while True:
            if ConfigObject.exists(git,
                                   project.get_section(),
                                   Worktree.Path.subsection(),
                                   str(path)):
                worktree = Worktree.get_by_path(git, project, str(path))
                break
            parent = path.parent
            if parent == path:
                break
            path = parent

    def add_arguments(self,
                      git,
                      gitproject,
                      project,
                      parser_manager,
                      plugin_manage):
        """Add arguments for 'git-project worktree.'"""

        # worktree
        worktree_parser = add_top_level_command(parser_manager,
                                                Worktree.get_managing_command(),
                                                Worktree.get_managing_command(),
                                                help='Manage worktrees',
                                                formatter_class=argparse.RawDescriptionHelpFormatter)

        add_plugin_version_argument(worktree_parser)

        worktree_subparser = parser_manager.add_subparser(worktree_parser,
                                                          'worktree-command',
                                                          help='worktree commands')

        # worktree add
        worktree_add_parser = parser_manager.add_parser(worktree_subparser,
                                                        'add',
                                                        'worktree-add',
                                                        help='Create a worktree',
                                                        epilog='One of path or committish is required.  If only one is specifiedd, the other will be inferred from the specified value.')

        worktree_add_parser.set_defaults(func=command_worktree_add)

        worktree_add_parser.add_argument('path',
                                         nargs='?',
                                         help='Path for worktree checkout')
        worktree_add_parser.add_argument('committish',
                                         nargs='?',
                                         help='Branch point for worktree')
        worktree_add_parser.add_argument('-b',
                                         '--branch',
                                         metavar='BRANCH',
                                         help='Create BRANCH for the worktree')

        # worktree rm
        worktree_rm_parser = parser_manager.add_parser(worktree_subparser,
                                                       'rm',
                                                       'worktree-rm',
                                                       help='Remove a worktree')

        worktree_rm_parser.set_defaults(func=command_worktree_rm)

        worktree_rm_parser.add_argument('name',
                                        help='Worktree to remove')
        worktree_rm_parser.add_argument('-f', '--force', action='store_true',
                                        help='Remove even if branch is not merged')

        # add a clone option to create a worktree layout.
        clone_parser = parser_manager.find_parser('clone')
        if clone_parser:
            clone_parser.add_argument('--worktree', action='store_true',
                                      help='Create a layout convenient for worktree use')

        # add an init option to create a worktree layout.
        init_parser = parser_manager.find_parser('init')
        if init_parser:
            init_parser.add_argument('--worktree', action='store_true',
                                     help='Create a layout convenient for worktree use')

    def _rewrite_bare_refspects(self, p_git):
        """Modify refspects to convert from a bare repository so that we merge origin
        branches to local branches.

        """
        assert p_git.is_bare_repository()

        p_git.set_remote_fetch_refspecs('origin',
                                        ['+refs/heads/*:refs/remotes/origin/*'])
        p_git.fetch_remote('origin')

        for refname in p_git.iterrefnames(['refs/heads']):
            if refname == 'refs/heads/master':
                remote_refname = p_git.get_remote_fetch_refname(refname, 'origin')
                p_git.set_branch_upstream(refname, remote_refname)
            else:
                p_git.delete_branch(refname)

    def _setup_master_worktree(self,
                               p_git,
                               p_gitproject,
                               p_project,
                               path,
                               clargs):
        """Create a master woorktree for a newly-created worktree layout."""
        # Set up a master worktree.
        master_path = path / 'master'
        setattr(clargs, 'committish', 'master')
        setattr(clargs, 'path', str(master_path))

        command_worktree_add(p_git, p_gitproject, p_project, clargs)

    def modify_arguments(self, git, gitproject, project, parser_manager, plugin_manager):
        """Modify arguments for 'git-project worktree.'"""

        # If a clone is done, set up a master worktree if told to.
        clone_parser = parser_manager.find_parser('clone')
        if clone_parser:
            command_clone = clone_parser.get_default('func')

            def worktree_command_clone(p_git, p_gitproject, p_project, clargs):
                if clargs.worktree:
                    path = Path.cwd()
                    if hasattr(clargs, 'path') and clargs.path:
                        path = Path(clargs.path)

                    path = path / '.git'

                    bare_specified = clargs.bare

                    setattr(clargs, 'path', str(path))
                    setattr(clargs, 'bare', True)

                    # Bare clone to the hidden .git directory.
                    path = command_clone(p_git, p_gitproject, p_project, clargs)

                    # If the user did not ask for a bare repo, rewrite refspecs,
                    # fetch remote refs and rewrite existing refs (just master)
                    # to track the remote ref.  Delete other "local" branches.
                    if not bare_specified:
                        self._rewrite_bare_refspects(p_git)

                    # Detach HEAD so we can worktree master.
                    p_git.detach_head()

                    self._setup_master_worktree(p_git,
                                                p_gitproject,
                                                p_project,
                                                Path(path).parent,
                                                clargs)
                else:
                    path = command_clone(p_git, p_gitproject, p_project, clargs)

                return path

            clone_parser.set_defaults(func=worktree_command_clone)

        # If an init is done, set up a master worktree layout if told to.
        init_parser = parser_manager.find_parser('init')
        if init_parser:
            command_init = init_parser.get_default('func')

            def worktree_command_init(p_git, p_gitproject, p_project, clargs):
                path = command_init(p_git, p_gitproject, p_project, clargs)

                if clargs.worktree:
                    was_bare = p_git.is_bare_repository()

                    # If it's not already, convert the current workarea to a bare repository.
                    if not p_git.is_bare_repository():
                        if not p_git.workarea_is_clean():
                            raise GitProjectException('Cannot initialize worktree layout, working copy not clean')

                        gitdir = Path(p_git.get_gitdir())
                        workarea_root = Path(p_git.get_working_copy_root())
                        assert workarea_root.exists()

                        if gitdir != workarea_root / '.git':
                            raise GitProjectException('Not creating worktree layout -- are you in a worktree?')

                        # Set bare and detach before removing files so they
                        # don't come back.  If we detach later, files will be
                        # checkout out.
                        p_git.config.set_item('core', 'bare', 'true')
                        p_git.detach_head()

                        # Remove everything except .git.
                        for filename in os.listdir(workarea_root):
                            if filename == '.git':
                                continue
                            path = workarea_root / filename
                            if os.path.isfile(path) or os.path.islink(path):
                                os.unlink(path)
                                assert not os.path.exists(path)
                            elif os.path.isdir(path):
                                shutil.rmtree(path)
                                assert not os.path.exists(path)

                    if was_bare:
                        workarea_root = Path(p_git.get_gitdir()).parent

                        # Note that since this is a bare repository, we may have
                        # branches besides master that are not pushed to
                        # whatever remote they should go to.  In general we
                        # cannot know which branches should go where so just
                        # punt and tell the user to clean them up.
                        for refname in p_git.iterrefnames(['refs/heads']):
                            if refname != 'refs/heads/master':
                                raise GitProjectException('Non-master branches detected, please push and/or delete them and try again.')

                        self._rewrite_bare_refspects(p_git)

                    self._setup_master_worktree(p_git,
                                                p_gitproject,
                                                p_project,
                                                workarea_root,
                                                clargs)

            init_parser.set_defaults(func=worktree_command_init)

    def iterclasses(self):
        """Iterate over public classes for git-project worktree."""
        yield Worktree



#!/usr/bin/env python3
#
# Copyright 2020 David A. Greene
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <https://www.gnu.org/licenses/>.
#

"""A plugin to add a 'clone' command to git-project.  The clone command clones a
repository and optionally initializes a master worktree environment.

Summary:

git-project clone <url> [path] [--bare] [--no-master-worktree]

"""

from git_project import add_top_level_command, Plugin

from git_project_core_plugins.common import add_plugin_version_argument

import getpass
import os
from pathlib import Path
import pygit2

def command_clone(git, gitproject, project, clargs):
    """Implement git-project clone"""
    gitdir = git.clone(clargs.url,
                       path=clargs.path if hasattr(clargs, 'path') else None,
                       bare=clargs.bare)

    # Now that we have a repository, add sensible project defaults.  We know
    # there is no existing project in the config file since we just cloned.
    # This will have the effect of adding a project section.
    project.set_defaults()

    return gitdir

class ClonePlugin(Plugin):
    """
    The clone command clones a repository.

    Summary:

      git <project> clone <url> [<path>] [--bare]

    By itself clone has just the very basic funcionality of the built-in git
    clone command.  Plugins may add options to give the clone command more
    features.  For example, the woktree command adds a --worktree option to have
    clone create a ``worktree layout.''

    See also:

      worktree

    """

    def __init__(self):
        super().__init__('clone')

    def add_arguments(self,
                      git,
                      gitproject,
                      project,
                      parser_manager,
                      plugin_manager):
        """Add arguments for 'git-project clone.'"""
        # clone
        clone_parser = add_top_level_command(parser_manager,
                                             'clone',
                                             'clone',
                                             help='Clone project')

        add_plugin_version_argument(clone_parser)

        clone_parser.set_defaults(func=command_clone)

        clone_parser.add_argument('url',
                                  help='URL to clone')

        clone_parser.add_argument('path',
                                  nargs='?',
                                  help='Local repository location')

        clone_parser.add_argument('--bare',
                                  action='store_true',
                                  help='Clone a bare repository')



#!/usr/bin/env python3
#
# Copyright 2020 David A. Greene
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <https://www.gnu.org/licenses/>.
#

"""A plugin to add a 'config' command to git-project.  The config command sets
project-wide git configuration values and prints their values to stdout.

Summary:

git-project config <key> [--unset] [<value>]

"""

from git_project import ConfigObject, RunnableConfigObject
from git_project import Plugin, Project, GitProjectException

from git_project_core_plugins.common import add_plugin_version_argument

def command_config(git, gitproject, project, clargs):
    """Implement git-project config."""
    getter = clargs.getter
    exister = clargs.exister
    classname = clargs.classname

    subsection = None
    if hasattr(clargs, 'subsection'):
        subsection = clargs.subsection

    ident = None
    if hasattr(clargs, 'ident'):
        ident = clargs.ident

    if not exister(git, project.get_section(), subsection, ident):
        if ident:
            raise GitProjectException(f'{classname} \'{ident}\' does not exist')
        else:
            raise GitProjectException(f'No {classname} configured')

    configitem = getter(git, project, ident) if ident else getter(git, project.get_section())

    if clargs.value:
        if clargs.unset:
            configitem.rm_item(clargs.name, clargs.value)
        else:
            if clargs.add:
                configitem.add_item(clargs.name, clargs.value)
            else:
                setattr(configitem, clargs.name, clargs.value)
    elif clargs.unset:
        delattr(configitem, clargs.name)
    else:
        value = getattr(configitem, clargs.name)
        print(value)

class ConfigPlugin(Plugin):
    """
    The config command manages git config settings under the <project> section.

    Summary:

      git <project> config [--add] [--unset] <name> [<value>]

    The config command operates much like git's built-in config command, except
    all configuration keys are prefixed with <project>, keeping values under a
    single project namespace.  This is a convenient way to store parameters for
    other commands.  For example:

      git <project> config builddir /path/to/build
      git <project> add run build "make BUILDDIR={builddir} all"

    Configuration kays may have their values substituted into other
    configuration values via the {key} specifier.  Special commands like build
    perform the substitution recursively, so configuration vaalues may contain
    substitutions of other configuration values which themselves contain
    substitutions, and so on.  Importantly, substitution only happens when
    commands are run.  Commands should document whether or not they perform
    substitutions.

    A git config ``sub-section'' may be substituted with its identifier.  For
    example:

      git <project> worktree add myworktree
      git <project> config builddir /path/to/{worktree}/build

    Here, myworktree is the identifier of a specific worktree sub-section.  If
    myworktree is the currently active worktree (that is, the current directory
    is under the myworktree root), then ``myworktree'' will substitute for
    {woktree}.

    See also:

      run
      worktree

    """
    def __init__(self):
        super().__init__('config')

    def _add_config_parser(self, cls, project, parser_manager):
        command = cls.get_managing_command()
        config_key = command + '-config' if command else 'config'
        config_parser = parser_manager.find_parser(config_key)

        if not config_parser:
            # Create a config parser under the managing command.
            command_subparser_key = command + '-command' if command else 'command'
            command_subparser = parser_manager.find_subparser(command_subparser_key)
            if not command_subparser:
                # This plugin doesn't have subcommands, so don't add one.
                return

            config_help = 'Configure ' + command if command else 'Configure git-project'
            config_parser = parser_manager.add_parser(command_subparser,
                                                      'config',
                                                      config_key,
                                                      help=config_help)

            config_parser.set_defaults(func=command_config)
            config_parser.set_defaults(getter=cls.get)
            config_parser.set_defaults(exister=cls.exists)
            config_parser.set_defaults(classname=cls.__name__)

            if hasattr(cls, 'subsection'):
                config_parser.set_defaults(subsection=cls.subsection())
                config_parser.add_argument('ident', help=f'{cls.__name__} to modify')

            if not command:
                # This is the top-level 'config' command.
                add_plugin_version_argument(config_parser)

            config_parser.add_argument('name', help='Property name')
            config_parser.add_argument('value', nargs='?', help='Property value to set')
            config_parser.add_argument('--add', action='store_true',
                                       help='Add a value to a property')
            config_parser.add_argument('--unset', action='store_true',
                                       help='Remove a value from a property')

    def add_arguments(self,
                      git,
                      gitproject,
                      project,
                      parser_manager,
                      plugin_manager):
        """Add arguments for 'git-project config'"""
        self._add_config_parser(Project, project, parser_manager)

    def modify_arguments(self, git, gitproject, project, parser_manager, plugin_manager):
        """Modify arguments for 'git-project config.'"""

        # Find all plugins with classes that derive from ConfigObject and add
        # config commands to them.
        for plugin in plugin_manager.iterplugins():
            for cls in plugin.iterclasses():
                if (issubclass(cls, ConfigObject)):
                    self._add_config_parser(cls, project, parser_manager)



#!/usr/bin/env python3
#
# Copyright 2020 David A. Greene
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <https://www.gnu.org/licenses/>.
#

"""A plugin to add a 'help' command to git-project.  The help command looks up
help text in the gitconfig and displays it.

Summary:

git-project help [command] [options...]

"""

from git_project import ConfigObject, Plugin, Project
from git_project import add_top_level_command, GitProjectException
from git_project import get_or_add_top_level_command

from git_project_core_plugins.common import add_plugin_version_argument

from pydoc import pager

def _print_general_help(git, gitproject, project, clargs):
    pager("""git-project: The stupid extensible project manager

git-project aims to make managing various project aspects simple and convenient.
The key idea of git-project is that various project actions may be coupled to
the source code repository or workarea in some fashion.  For example, developers
may want builds isolated by the current branch, such that build directories are
encoded with the branch name.  Or users may want to create worktrees associated
with build and install directories, so that removing a worktree also removes the
associated build and install directories.

git-project may be invoked via ``git project <command> [options...]'' or, by
linking the git-project command to another command named ``git-<name>'' it may
be invoked via ``git <name> <command>.''  For example, linking ``git-project''
to ``git-fizzbin'' we may invoke commands via ``git fizzbin <command>'' and all
configuration will appear under a ``fizzbin'' config section.  Thus, a single
repository may support multiple projects, each with a unique name.  To emphasize
this, we will show invocations of git-project as ``git <project> [options...].''

Plugins extend git-project with new functionality.  Each plugin should include
extensive documentation accessible though --help flags or via help command
options.

To see help for a specific command, try ``git <project> help [command]
 [options...].''
""")

class Help(ConfigObject):
    @staticmethod
    def subsection():
        """ConfigObject protocol subsection."""
        return 'help'

    @classmethod
    def get_managing_command(cls):
        """ConfigObject protocol get_managing_command."""
        return 'help'

    @staticmethod
    def _prepend_subsection(subsection):
        if subsection:
            return 'help.' + subsection
        return 'help'

    @classmethod
    def get(cls, git, project, subsection, ident, **kwargs):
        """Factory to construct Helps.

        cls: The derived class being constructed.

        git: An object to query the repository and make config changes.

        project: The currently active Project.

        subsection: The subsection under "help" to retrieve.

        ident: The specific help to retrieve.

        kwargs: Attributes to set.

        """
        return super().get(git,
                           project.get_section(),
                           cls._prepend_subsection(subsection),
                           ident,
                           **kwargs)

    @classmethod
    def exists(cls,
               git,
               project,
               subsection,
               ident):
        """Return whether an existing git config exists for the ConfigObject.

        cls: The derived class being checked.

        git: An object to query the repository and make config changes.

        project: The active project.

        subsection: Subsection under 'help' to query

        ident: The name of the specific help we're looking for.

        """
        return super().exists(git,
                              project.get_section(),
                              cls._prepend_subsection(subsection),
                              ident)

# Map command names to plugins.  We use this to see if a command given to
# git-project help has its own built-in help in its plugin.  This avoids the
# need to store large help pages in the git config and also allows later
# versions of plugins to provide new help and not be stuck with whatever happens
# to be stored in the config.
_command_plugin_registry = dict()

def command_help(git, gitproject, project, clargs):
    """Implement git-project help."""
    if not hasattr(clargs, 'options') or len(clargs.options) < 1:
        _print_general_help(git, gitproject, project, clargs)
        return

    ident = clargs.options[-1]
    subsection = ''

    if len(clargs.options) == 1:
        # See if this command has built-in help.
        plugin = _command_plugin_registry.get(ident, None)
        if hasattr(plugin, 'manpage'):
            pager(plugin.manpage())
            return
    else:
        subsection = '.'.join(clargs.options[:-1])

    if not Help.exists(git, project, subsection, ident):
        print(f'No help available for \'{" ".join(clargs.options)}\'')
        return

    command_help = Help.get(git,project, subsection, ident)

    if (hasattr(command_help, 'manpage')):
        pager(command_help.manpage)
    else:
        print(f'No help page available for \'{" ".join(clargs.options)}\'')

def command_add_help(git, gitproject, project, clargs):
    f"""Implement git-project add help"""
    help_section = f'{project.get_section()}.help.{clargs.name}'
    if clargs.manpage:
        git.config.add_item(help_section, 'manpage', clargs.text)
    else:
        git.config.add_item(help_section, 'short', clargs.text)

def command_rm_help(git, gitproject, project, clargs):
    f"""Implement git-project r help"""
    help_section = f'{project.get_section()}.help.{clargs.name}'
    if clargs.manpage:
        git.config.rm_items(help_section, 'manpage')
    else:
        git.config.rm_items(help_section, 'short')

class HelpPlugin(Plugin):
    """
    The help command displays tutorial-style help for commands.

    Summary:

      git <project> add help [--manpage] <subsection> <text>
      git <project> help <command>

    Users may add help to any project config section.  For example:

      git <project> add help run.build "Perform a build"
      git <project> add help run.check "Run tests"

    All help is stored under a <project>.help config sub-section.  If a command
    supports it, such help may appear in the command's own help output by
    querying the appropriate <project>.help sub-section:

      git <project> run --help

      <standard help text>

      build     -- Perform a build
      check     -- Run tests

    In this way projects can self-document their configurations.  Normally
    <text> is stored in <project>.help.<subsection>.short.  With --manpage,
    <text> is stored in <project>.help.<subsection>.manpage.  Commands may
    reference short help or manpages in various ways to present help.

    See also:

      run

    """
    def __init__(self):
        super().__init__('help')

    def add_arguments(self,
                      git,
                      gitproject,
                      project,
                      parser_manager,
                      plugin_manager):
        """Add arguments for 'git-project help'"""
        # help
        help_parser = add_top_level_command(parser_manager,
                                            'help',
                                            'help',
                                            help='Display help')

        help_parser.set_defaults(func=command_help)

        add_plugin_version_argument(help_parser)

        help_parser.add_argument('options',
                                 nargs='*',
                                 help='Specify help section')

        # add help
        add_parser = get_or_add_top_level_command(parser_manager,
                                                  'add',
                                                  'add',
                                                  help=f'Add config sections to {project.get_section()}')

        add_subparser = parser_manager.get_or_add_subparser(add_parser,
                                                            'add-command',
                                                            help='add sections')

        add_help_parser = parser_manager.add_parser(add_subparser,
                                                    'help',
                                                    'add-help',
                                                    help=f'Add help to {project.get_section()}')

        add_help_parser.set_defaults(func=command_add_help)

        add_help_parser.add_argument('--manpage',
                                     action='store_true',
                                     help='Add help as manpage')

        add_help_parser.add_argument('name',
                                     help='Command name for which to add help')

        add_help_parser.add_argument('text',
                                     help='Help text')

        # rm help
        rm_parser = get_or_add_top_level_command(parser_manager,
                                                  'rm',
                                                  'rm',
                                                  help=f'Remove config sections from {project.get_section()}')

        rm_subparser = parser_manager.get_or_add_subparser(rm_parser,
                                                           'rm-command',
                                                           help='remove sections')

        rm_help_parser = parser_manager.add_parser(rm_subparser,
                                                   'help',
                                                   'rm-help',
                                                   help=f'Remove help from {project.get_section()}')

        rm_help_parser.set_defaults(func=command_rm_help)

        rm_help_parser.add_argument('--manpage',
                                    action='store_true',
                                    help='Add help as manpage')

        rm_help_parser.add_argument('name',
                                    help='Command name for which to add help')

    def modify_arguments(self,
                         git,
                         gitproject,
                         project,
                         parser_manager,
                         plugin_manager):
        """Register known plugins for 'git-project help.'"""

        for plugin in plugin_manager.iterplugins():
            _command_plugin_registry[plugin.name] = plugin



#!/usr/bin/env python3
#
# Copyright 2020 David A. Greene
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <https://www.gnu.org/licenses/>.
#

"""A plugin to add a 'branch' command to git-project.  The branch command allows
query of branch status with respect to the project and pruning of 'finished'
branches.

Summary:

git-project branch status [<pattern>]
git-project branch prune [--no-ask] [<pattern>]

"""
from git_project import Git, RunnableConfigObject, Plugin
from git_project import add_top_level_command, Project, GitProjectException

from git_project_core_plugins.common import add_plugin_version_argument

import getpass

def query_yes_no(question, default="yes"):
    """Ask a yes/no question via input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")

def command_branch_status(git, gitproject, project, clargs):
    """Implement git-project branch status."""
    ref = clargs.name_or_ref

    if not ref:
        if clargs.all_user:
            ref = 'refs/heads/user/' + getpass.getuser()
        elif clargs.all:
            ref = 'refs/heads'
        else:
            raise GitProjectException('merged requires branch name, pattern, --all-user or --all')
    else:
        if not ref.startswith('refs/'):
            ref = 'refs/heads/' + ref

    target = clargs.target if clargs.target else None

    branch_width = 45
    status_width = 8
    separator_width = branch_width + 2 * status_width - (status_width - len('merged'))

    print('-' * separator_width)
    print('{:<{branch_width}s}{:<{status_width}s}{:<{status_width}s}'.
          format('branch', 'merged', 'pushed',
                 branch_width=branch_width, status_width=status_width))
    print('-' * separator_width)

    for branch in git.iterrefnames([ref]):
        merge_status = 'no'
        if target:
            merge_status = 'yes' if git.branch_is_merged(branch, target) else 'no'
        else:
            merge_status = 'yes' if project.branch_is_merged(branch) else 'no'

        push_status = 'yes' if project.branch_is_pushed(branch) else 'no'

        print('{:<{branch_width}s}{:<{status_width}s}{:<{status_width}s}'.
              format(branch[:branch_width-2], merge_status, push_status,
                     branch_width=branch_width, status_width=status_width))
    print('-' * separator_width)

def command_branch_prune(git, gitproject, project, clargs):
    """Implement git-project branch prune."""
    ref = clargs.name_or_ref
    if not ref:
        if clargs.all_user:
            ref = 'refs/heads/user/' + getpass.getuser()
        else:
            raise GitProjectException('Prune requires branch name, pattern or --all-user')
    else:
        ref = Git.branch_name_to_refname(ref)

    branch_width = 45
    status_width = 15
    separator_width = branch_width + 2 * status_width

    print('-' * separator_width)
    print('{:<{branch_width}s}{:<{status_width}s}{:<{status_width}s}'.
          format('branch', 'local status', 'remote status',
                 branch_width=branch_width, status_width=status_width))
    print('-' * separator_width)

    for branch in git.iterrefnames([ref]):
        status = 'merged' if project.branch_is_merged(branch) else 'unmerged'

        print('{:<{branch_width}s}{:<{status_width}s}'.
              format(branch[:branch_width-2], status,
                     branch_width=branch_width, status_width=status_width))
        if clargs.force or status == 'merged':
            if clargs.no_ask or query_yes_no('Prune?', default=None):
                project.prune_branch(branch)

class BranchPlugin(Plugin):
    """
    The branch command queries the status of branches against known project
    branches and provides methods to prune old branches.

    Summary:

      git <project> branch status [--all] [<refish>]
      git <project> branch prune [--force] [--no-ask]

    The branch status command checks the given <refish> (or all local branches
    with the --all option) against the project-configured branches.  The command
    outputs a table of branches and whether they are merged to a project branch
    and/or pushed to a remote.  For example:

      git <project> config --add branch release
      git <project> branch status mybranch

    The report will show whether mybranch is merged to the release or master
    branches (master is always a configured project branch) and whether the
    commit pointed to mybranch is pushed to a remote.

    The branch prune command computes the same information and if the branch is
    merged to a project branch and that project branch is pushed to a remote,
    will ask whether mybranch should be deleted.  If the user indicates yes,
    both the local mybranch and its remote counterpart, if any, will be deleted.

    With --force, branches will be pruneed regardless of merge/push status.
    With --no-ask branch prune operates in batch mode, assuming all merged and
    pushed branches should be pruned.

    See also:

      config

    """
    def __init__(self):
        super().__init__('branch')

    def add_arguments(self,
                      git,
                      gitproject,
                      project,
                      parser_manager,
                      plugin_manager):
        """Add arguments for 'git project branch.'"""
        # branch
        branch_parser = add_top_level_command(parser_manager,
                                              'branch',
                                              'branch',
                                              help='Manipulate branches')

        add_plugin_version_argument(branch_parser)

        branch_subparser = parser_manager.add_subparser(branch_parser,
                                                        'branch_command',
                                                        help='branch commands')

        # branch status
        branch_status_parser = parser_manager.add_parser(branch_subparser,
                                                         'status',
                                                         'branch-status',
                                                         help='Show branch status')

        branch_status_parser.set_defaults(func=command_branch_status)

        branch_status_parser.add_argument('name_or_ref', nargs='?', help='Branch or pattern to filter branches')
        branch_status_parser.add_argument('target', nargs='?', help='Target branch to check against')
        branch_status_parser.add_argument('--all', action='store_true', help='Show all branches')
        branch_status_parser.add_argument('--all-user', action='store_true', help='Show all user\'s branches')

        # branch prune
        branch_prune_parser = parser_manager.add_parser(branch_subparser,
                                                        'prune',
                                                        'branch-prune',
                                                        help='Delete merged branches')

        branch_prune_parser.set_defaults(func=command_branch_prune)

        branch_prune_parser.add_argument('name_or_ref', nargs='?', help='Branch name or pattern to filter branches')
        branch_prune_parser.add_argument('--all-user', action='store_true', help='Prune all user\'s branches')
        branch_prune_parser.add_argument('--force', action='store_true', help='Prune even if unmerged')
        branch_prune_parser.add_argument('--no-ask', action='store_true', help='Do not ask before pruning')



#!/usr/bin/env python3
#
# Copyright 2020 David A. Greene
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <https://www.gnu.org/licenses/>.
#

from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()

setup(name='git_project_core_plugins',
      version='0.0.14',
      description='Core functionality for git-project',
      long_description=readme(),
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
          'Programming Language :: Python :: 3.6',
          'Topic :: Software Development :: Version Control :: Git',
      ],
      keywords='git project development',
      url='http://github.com/greened/git-project-core-plugins',
      author='David A. Greene',
      author_email='dag@obbligato.org',
      license='GPLv3+',
      packages=['git_project_core_plugins'],
      install_requires = [
          'git_project>=0.0.12',
          'progressbar2',
          'pygit2',
      ],
      python_requires='~=3.8',
      entry_points = {
          'git_project.plugins': {
              'artifact = git_project_core_plugins.artifact:ArtifactPlugin',
              'branch = git_project_core_plugins.branch:BranchPlugin',
              'run = git_project_core_plugins.run:RunPlugin',
              'clone = git_project_core_plugins.clone:ClonePlugin',
              'config = git_project_core_plugins.config:ConfigPlugin',
              'help = git_project_core_plugins.help:HelpPlugin',
              'init = git_project_core_plugins.init:InitPlugin',
              'worktree = git_project_core_plugins.worktree:WorktreePlugin',
          }
      },
      zip_safe=False)



#!/usr/bin/env python3
#
# Copyright 2020 David A. Greene
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <https://www.gnu.org/licenses/>.
#

"""A plugin to add a 'run' command to git-project.  The run command invokes an
arbitrary command.  It can be used to perform any action, such as building the
project.

Summary:

git-project run <name>

"""

from git_project import ConfigObject, RunnableConfigObject, Plugin, Project
from git_project import get_or_add_top_level_command, GitProjectException

from git_project_core_plugins.common import add_plugin_version_argument

import argparse

class RunConfig(ConfigObject):
    """A ConfigObject to manage run aliases."""

    @staticmethod
    def subsection():
        """ConfigObject protocol subsection."""
        return 'run'

    def __init__(self,
                 git,
                 project_section,
                 subsection,
                 ident = None,
                 **kwargs):
        """RunConfig construction.

        cls: The derived class being constructed.

        git: An object to query the repository and make config changes.

        project_section: git config section of the active project.

        subsection: An arbitrarily-long subsection appended to project_section

        **kwargs: Keyword arguments of property values to set upon construction.

        """
        super().__init__(git,
                         project_section,
                         subsection,
                         ident,
                         **kwargs)

    @classmethod
    def get(cls, git, project, **kwargs):
        """Factory to construct RunConfigs.

        cls: The derived class being constructed.

        git: An object to query the repository and make config changes.

        project: The currently active Project.

        kwargs: Attributes to set.

        """
        return super().get(git,
                           project.get_section(),
                           cls.subsection(),
                           None,
                           **kwargs)

class RunPlugin(Plugin):
    """
    The run command executes commands via a shell.

    Summary:

      git <project> add run <name> <command>
      git <project> run --make-alias <name>
      git <project> run <name>

    Full shell substitution is supported, as well as config {key} substitution,
    where the text ``{key}'' is replaced by key's value.

    The add run command associates a command string with a name.  The run
    command itself invokes the command string via a shell.  With --make-alias,
    the run comand instead registers an alternative name for ``run.''  For
    example:

      git <project> run --make-alias build
      git <project> add build all "make -C {path} all"
      git <project> build all

    Note that an alias will prevent ``run'' from invoking the command so in the
    above example we could not invoke the build as such:

      git <project> run all

    In this way we may use the same <name> for different registered run aliases,
    which can be convenient:

      git <project> build all
      git <project> check all

    In general any project config {key} will be replaced with its value before
    the command is executed.  Substitution occurs recursively, so if a {key}
    value itself contains a substitution string, it will be replaced as well.
    There are a few special case substitutions.  The {path} key will be replaced
    by the absolute path to the root of the current workarea.  The {branch} key
    will be replaced by the currently-active branch nme.  In addition the {run}
    (or {build}, etc. aliases) will be replaced by their names.  Again, an
    example will make this more clear:

      git <project> config cmd "make -C {path} BLDDIR=/path/to/{build} {build}"
      git project add build all "{cmd}"
      git project add build some "{cmd}"

    We have configured two different build flavors, each which place build
    results in separate directories and invoke different targets.  Substitution
    proceeds as follows:

      git project build all -> make -C /cur/workarea BLDDIR=/path/to/all all
      git project build some -> make -C /cur/workarea BLDDIR=/path/to/some some

    Some plugins may add scoping rules to the project config, such that a scope
    nested inside the project may override the global project config key value.
    For example the worktree plugin adds a ``worktree'' scope.  The worktree may
    contain key values that override similar keys in the project config.

    See also:

      config
      worktree

    """
    def __init__(self):
        super().__init__('run')
        self.classes = dict()
        self.classes['run'] = self._make_alias_class('run')

    def _make_alias_class(self, alias):
        # Create a class for the alias.
        @staticmethod
        def subsection():
            """ConfigObject protocol subsection."""
            return alias

        @classmethod
        def get_managing_command(cls):
            """ConfigObject protocol get_managing_command."""
            return alias

        Class = type(alias + "Class", (RunnableConfigObject, ), {
#            __doc__ = f"""A RunnableConfigObject to manage {alias} names.  Each run name gets its own
#            config section.
#
#            """
                'subsection': subsection,
                'get_managing_command': get_managing_command
        })

        def cons(self,
                 git,
                 project_section,
                 subsection,
                 ident,
                 **kwargs):
            f"""{alias} construction.

            cls: The derived class being constructed.

            git: An object to query the repository and make config changes.

            project_section: git config section of the active project.

            subsection: An arbitrarily-long subsection appended to project_section

            ident: The name of this specific {alias}.

            **kwargs: Keyword arguments of property values to set upon construction.

            """
            super(Class, self).__init__(git,
                                        project_section,
                                        subsection,
                                        ident,
                                        **kwargs)

        @classmethod
        def get(cls, git, project, name, **kwargs):
            f"""Factory to construct {alias}s.

            cls: The derived class being constructed.

            git: An object to query the repository and make config changes.

            project: The currently active Project.

            name: Name of the command to run.

            kwargs: Attributes to set.

            """
            return super(Class, cls).get(git,
                                         project.get_section(),
                                         cls.subsection(),
                                         name,
                                         **kwargs)

        Class.__init__ = cons
        Class.get = get

        self.classes[alias] = Class
        return Class

    def _gen_runs_epilog(self, git, project, alias, runs):
        result = f'Available {alias}s:\n'
        run_width = 20
        for run in runs:
            help_section = f'{project.get_section()}.help.{alias}.{run}'
            help_key = 'short'
            if git.config.has_item(help_section, help_key):
                shorthelp = git.config.get_item(help_section, help_key)
                result += f'    {run:<{run_width}} - {shorthelp}\n'
            else:
                result += f'    {run}\n'

        return result

    def _add_alias_arguments(self,
                             git,
                             gitproject,
                             project,
                             parser_manager,
                             Class):
        alias = Class.get_managing_command()

        # add run
        add_parser = get_or_add_top_level_command(parser_manager,
                                                  'add',
                                                  'add',
                                                  help=f'Add config sections to {project.get_section()}')

        add_subparser = parser_manager.get_or_add_subparser(add_parser,
                                                            'add-command',
                                                            help='add sections')

        add_run_parser = parser_manager.add_parser(add_subparser,
                                                   alias,
                                                   'add-' + alias,
                                                   help=f'Add a {alias} to {project.get_section()}')

        def command_add_run(git, gitproject, project, clargs):
            f"""Implement git-project add {alias}"""
            run = Class.get(git,
                            project,
                            clargs.name,
                            command=clargs.command)
            project.add_item(alias, clargs.name)
            return run


        add_run_parser.set_defaults(func=command_add_run)

        add_run_parser.add_argument('name',
                                    help='Name for the run')

        add_run_parser.add_argument('command',
                                    help='Command to run')

        runs = []
        if hasattr(project, alias):
            runs = [run for run in project.iter_multival(alias)]

        # rm run
        rm_parser = get_or_add_top_level_command(parser_manager,
                                                 'rm',
                                                 'rm',
                                                 help=f'Remove config sections from {project.get_section()}')

        rm_subparser = parser_manager.get_or_add_subparser(rm_parser,
                                                           'rm-command',
                                                           help='rm sections')

        rm_run_parser = parser_manager.add_parser(rm_subparser,
                                                  alias,
                                                  'rm-' + alias,
                                                  help=f'Remove a {alias} from {project.get_section()}')

        def command_rm_run(git, gitproject, project, clargs):
            f"""Implement git-project rm {alias}"""
            run = Run.get(git, project, alias, clargs.name, command=clargs.command)
            run.rm()
            print(f'Removing project {alias} {clargs.name}')
            project.rm_item(alias, clargs.name)

        rm_run_parser.set_defaults(func=command_rm_run)

        if runs:
            rm_run_parser.add_argument('name', choices=runs,
                                       help='Command name')

        # run
        command_subparser = parser_manager.find_subparser('command')

        run_parser = parser_manager.add_parser(command_subparser,
                                               alias,
                                               alias,
                                               help=f'Invoke {alias}',
                                               epilog=self._gen_runs_epilog(git,
                                                                            project,
                                                                            alias,
                                                                            runs),
                                               formatter_class=
                                               argparse.RawDescriptionHelpFormatter)

        def command_run(git, gitproject, project, clargs):
            """Implement git-project run"""
            if clargs.make_alias:
                run_config = RunConfig.get(git, project)
                run_config.add_item('alias', clargs.name)
            else:
                if not clargs.name in runs:
                    raise GitProjectException(f'Unknown {alias} "{clargs.name}," choose one of: {{ {runs} }}')
                run = Class.get(git, project, clargs.name)

                translation_table = dict.fromkeys(map(ord, '{}'), None)

                option_names = ' '.join(clargs.options)
                option_names = option_names.translate(translation_table)
                option_key = '-'.join(clargs.options)
                option_key = option_key.translate(translation_table)

                formats = {
                    'options': ' '.join(clargs.options),
                    'option_names': option_names,
                    'option_key': option_key,
                    'option_keysep': '-' if len(clargs.options) > 0 else ''
                }

                run.run(git, project, formats)

        run_parser.set_defaults(func=command_run)

        add_plugin_version_argument(run_parser)

        run_parser.add_argument('--make-alias', action='store_true',
                                help=f'Alias "{alias}" to another command')

        run_parser.add_argument('name', help='Command name or alias')

        run_parser.add_argument('options',
                                nargs='*',
                                help='Additions options to pass to command')

    def add_arguments(self,
                      git,
                      gitproject,
                      project,
                      parser_manager,
                      plugin_manage):
        """Add arguments for 'git-project run.'"""
        if git.has_repo():
            # Get the global run ConfigObject and add any aliases.
            run_config = RunConfig.get(git, project)
            for alias in run_config.iter_multival('alias'):

                Class = self._make_alias_class(alias)

            for Class in self.iterclasses():
                self._add_alias_arguments(git,
                                          gitproject,
                                          project,
                                          parser_manager,
                                          Class)

    def get_class_for(self, alias):
        return self.classes[alias]

    def iterclasses(self):
        """Iterate over public classes for git-project run."""
        for key, Class in self.classes.items():
            yield Class
